import base64
import urllib.parse
from typing import Optional
from itertools import cycle

class Encryption:
    def __init__(self, key: Optional[bytes] = None):
        self.key = key if key else b'ali1343faraz1055antler288based'

    def encrypt(self, message: str) -> str:
        """Encrypts a message using XOR encryption."""
        message_bytes = message.encode('utf-8')
        output_bytes = bytes(m ^ k for m, k in zip(message_bytes, cycle(self.key)))
        encoded = base64.b64encode(output_bytes).decode('utf-8')
        return urllib.parse.quote(encoded)

    def decrypt(self, encrypted: str) -> str:
        """Decrypts an encrypted message using XOR decryption."""
        try:
            unquoted = urllib.parse.unquote(encrypted)
            if encrypted.startswith("{"):
                return encrypted
            encrypted_bytes = base64.b64decode(unquoted)
        except (ValueError, base64.binascii.Error) as err:
            print(f"Error during decoding: {err}")
            return ""

        decrypted_bytes = bytes(e ^ k for e, k in zip(encrypted_bytes, cycle(self.key)))
        return decrypted_bytes.decode('utf-8', errors='ignore')



class FruitCraftException(Exception):
    """Base class for all FruitCraft exceptions."""

    def __init__(self, message, code=None, params: tuple = ()):
        super().__init__(message)
        self.code = code
        self.params = params


class UnknownError(FruitCraftException):
    def __init__(self):
        super().__init__("An unknown error occurred.")


class GeneralError(FruitCraftException):
    """Raised for general errors."""

    def __init__(self, params: tuple = (), message: str = "An unexpected error occurred, please try again.",
                 code: int = 100):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AccountBlocked(FruitCraftException):
    """Raised when the account is blocked."""

    def __init__(self, params: tuple = (), message: str = "Your account is blocked.", code: int = 101):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardNotFound(FruitCraftException):
    """Raised when a card is not found."""

    def __init__(self, params: tuple = (), message: str = "Card not found.", code: int = 102):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NoCardSelected(FruitCraftException):
    """Raised when no card is selected."""

    def __init__(self, params: tuple = (), message: str = "You should choose at least one card!", code: int = 103):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardAlreadyInAuction(FruitCraftException):
    """Raised when a card is already in auction."""

    def __init__(self, params: tuple = (), message: str = "Card is already in auction.", code: int = 104):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotEnoughCards(FruitCraftException):
    """Raised when there are not enough cards."""

    def __init__(self, params: tuple = (), message: str = "Not enough cards.", code: int = 105):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InconsistencyError(FruitCraftException):
    """Raised for inconsistency errors."""

    def __init__(self, params: tuple = (),
                 message: str = "An inconsistency error occurred about your action, please try again.",
                 code: int = 106):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardNotInAuction(FruitCraftException):
    """Raised when a card is not found in auctions."""

    def __init__(self, params: tuple = (), message: str = "Card not found in auctions.", code: int = 107):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxBidReached(FruitCraftException):
    """Raised when the maximum bid is reached."""

    def __init__(self, params: tuple = (),
                 message: str = "You cannot bid on this item, it's already reached the maximum price.",
                 code: int = 108):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CannotBidOwnCard(FruitCraftException):
    """Raised when a user tries to bid on their own card."""

    def __init__(self, params: tuple = (), message: str = "You cannot bid on your own cards.", code: int = 109):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AlreadyHighestBidder(FruitCraftException):
    """Raised when the user is already the highest bidder."""

    def __init__(self, params: tuple = (), message: str = "You are already the highest bidder.", code: int = 110):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AuctionClosed(FruitCraftException):
    """Raised when the auction is closed."""

    def __init__(self, params: tuple = (), message: str = "Too late! Auction is closed.", code: int = 111):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardTypeQuery(FruitCraftException):
    """Raised when querying for a card type."""

    def __init__(self, params: tuple = (), message: str = "What kind of card should I look for?", code: int = 112):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardTypeNotFound(FruitCraftException):
    """Raised when the selected card type is not found."""

    def __init__(self, params: tuple = (), message: str = "We cannot find the card type selected.", code: int = 113):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardTypeQueryDuplicate(FruitCraftException):
    """Raised when querying for a card type is duplicated."""

    def __init__(self, params: tuple = (), message: str = "What kind of card should I look for?", code: int = 114):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotOwnerError(FruitCraftException):
    """Raised when the user is not the owner of a card."""

    def __init__(self, params: tuple = (), message: str = "You are not the owner. Cannot proceed.", code: int = 115):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AccessDenied(FruitCraftException):
    """Raised when access is denied."""

    def __init__(self, params: tuple = (), message: str = "Access denied.", code: int = 116):
        super().__init__(message % params)
        self.code = code
        self.params = params


class ServerMaintenance(FruitCraftException):
    """Raised when the server is under maintenance."""

    def __init__(self, params: tuple = (), message: str = "Server is in maintenance. We will be available soon.",
                 code: int = 117):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NoCardsSpecified(FruitCraftException):
    """Raised when no cards are specified for an operation."""

    def __init__(self, params: tuple = (), message: str = "You need to specify some cards for this operation.",
                 code: int = 118):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NoOpponentSpecified(FruitCraftException):
    """Raised when no opponent is specified for an operation."""

    def __init__(self, params: tuple = (), message: str = "You need to specify an opponent for this operation.",
                 code: int = 119):
        super().__init__(message % params)
        self.code = code
        self.params = params


class BattleNotAvailable(FruitCraftException):
    """Raised when a battle cannot be started."""

    def __init__(self, params: tuple = (), message: str = "You can't start the battle right now, please try again.",
                 code: int = 120):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CannotAttackSelf(FruitCraftException):
    """Raised when a user tries to attack themselves."""

    def __init__(self, params: tuple = (), message: str = "You cannot attack yourself.", code: int = 121):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerProtected(FruitCraftException):
    """Raised when a player is protected by a shield."""

    def __init__(self, params: tuple = (),
                 message: str = "Player is protected by shield, wait a few hours before attacking them again.",
                 code: int = 122):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CaptchaRequired(FruitCraftException):
    """Raised when a captcha code is required to proceed."""

    def __init__(self, params: tuple = (), message: str = "You have to enter captcha code to proceed.",
                 code: int = 123):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayingOnAnotherDevice(FruitCraftException):
    """Raised when the user is playing on another device."""

    def __init__(self, params: tuple = (),
                 message: str = "You are currently playing on another device, please close the game on other devices and try again in a few seconds.",
                 code: int = 124):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OpponentNotFound(FruitCraftException):
    """Raised when the opponent is not found."""

    def __init__(self, params: tuple = (), message: str = "Opponent not found.", code: int = 125):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OpponentOutOfRange(FruitCraftException):
    """Raised when the opponent is out of range."""

    def __init__(self, params: tuple = (), message: str = "Opponent not in your range.", code: int = 126):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OpponentDefenseEmpty(FruitCraftException):
    """Raised when the opponent's defense deck is empty."""

    def __init__(self, params: tuple = (), message: str = "Opponent's defence deck is empty.", code: int = 127):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidCaptcha(FruitCraftException):
    """Raised when the captcha is invalid."""

    def __init__(self, params: tuple = (), message: str = "Captcha is invalid.", code: int = 128):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardInUse(FruitCraftException):
    """Raised when a card is in use and cannot be used in another operation."""

    def __init__(self, params: tuple = (),
                 message: str = "Card is used in one of your buildings and cannot be used here.", code: int = 129):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardMarkedForSacrifice(FruitCraftException):
    """Raised when a card is already marked for sacrifice."""

    def __init__(self, params: tuple = (), message: str = "Card is already marked for sacrifice.", code: int = 130):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxPowerReached(FruitCraftException):
    """Raised when a card has reached its maximum power."""

    def __init__(self, params: tuple = (), message: str = "Card has reached its maximum power.", code: int = 131):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxEvolutionsExceeded(FruitCraftException):
    """Raised when more than two cards are attempted to be evolved."""

    def __init__(self, params: tuple = (), message: str = "Cannot evolve more than two cards.", code: int = 132):
        super().__init__(message % params)
        self.code


class EvolveLimitExceeded(FruitCraftException):
    """Raised when trying to evolve more than two cards."""

    def __init__(self, params: tuple = (), message: str = "Only two cards of the same type can be evolved.",
                 code: int = 133):
        super().__init__(message % params)
        self.code = code
        self.params = params


class BuildingNotFound(FruitCraftException):
    """Raised when a building is not found."""

    def __init__(self, params: tuple = (), message: str = "The specified building was not found.", code: int = 134):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxCardsAssigned(FruitCraftException):
    """Raised when the maximum number of cards has been assigned."""

    def __init__(self, params: tuple = (), message: str = "Maximum number of cards assigned.", code: int = 135):
        super().__init__(message % params)
        self.code = code
        self.params = params


class LiveBattleUnavailable(FruitCraftException):
    """Raised when live battle is not available."""

    def __init__(self, params: tuple = (),
                 message: str = "Live-battle is not available due to some inconsistencies. We will be available soon.",
                 code: int = 136):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OpponentOffline(FruitCraftException):
    """Raised when the opponent is no longer online."""

    def __init__(self, params: tuple = (), message: str = "Opponent is no longer online.", code: int = 137):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AlreadyInLiveBattle(FruitCraftException):
    """Raised when the user is already in a live battle."""

    def __init__(self, params: tuple = (), message: str = "Already in live-battle.", code: int = 138):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OpponentBusy(FruitCraftException):
    """Raised when the opponent is busy."""

    def __init__(self, params: tuple = (), message: str = "Opponent is busy.", code: int = 139):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CannotAttackTribeMate(FruitCraftException):
    """Raised when a user tries to attack their tribemate."""

    def __init__(self, params: tuple = (), message: str = "You cannot attack your tribemate.", code: int = 140):
        super().__init__(message % params)
        self.code = code
        self.params = params


class BattleIdRequired(FruitCraftException):
    """Raised when a battle ID is required."""

    def __init__(self, params: tuple = (), message: str = "Battle ID required.", code: int = 141):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InconsistencyErrorUpdate(FruitCraftException):
    """Raised for inconsistency errors during updates."""

    def __init__(self, params: tuple = (),
                 message: str = "An inconsistency error occurred. Please update your game to the latest version and try again.",
                 code: int = 142):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OperationTimeout(FruitCraftException):
    """Raised when an operation times out."""

    def __init__(self, params: tuple = (), message: str = "The operation is timed out. Please try again.",
                 code: int = 143):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerUnavailable(FruitCraftException):
    """Raised when a player is not available."""

    def __init__(self, params: tuple = (),
                 message: str = "Player is not available right now. Please try again in a moment.", code: int = 144):
        super().__init__(message % params)
        self.code = code
        self.params = params


class BattleNotFound(FruitCraftException):
    """Raised when a battle is not found."""

    def __init__(self, params: tuple = (), message: str = "Battle not found. Please try again.", code: int = 145):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AttackerNotFound(FruitCraftException):
    """Raised when the attacker is not found."""

    def __init__(self, params: tuple = (), message: str = "Attacker not found. Please try again.", code: int = 146):
        super().__init__(message % params)
        self.code = code
        self.params = params


class DefenderNotFound(FruitCraftException):
    """Raised when the defender is not found."""

    def __init__(self, params: tuple = (), message: str = "Defender not found. Please try again.", code: int = 147):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeNotFound(FruitCraftException):
    """Raised when a tribe is not found."""

    def __init__(self, params: tuple = (), message: str = "Tribe not found. Please try again.", code: int = 148):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotInTribe(FruitCraftException):
    """Raised when the user is not in a tribe."""

    def __init__(self, params: tuple = (), message: str = "You are not in a tribe.", code: int = 149):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeHelpLimit(FruitCraftException):
    """Raised when the tribe cannot use help due to limitations."""

    def __init__(self, params: tuple = (),
                 message: str = "Your tribe cannot use help for not conforming to the tribe member limitation rule.",
                 code: int = 150):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MessageSizeExceeded(FruitCraftException):
    """Raised when the message size exceeds the limit."""

    def __init__(self, params: tuple = (), message: str = "Message size exceeded.", code: int = 151):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MessageNotFound(FruitCraftException):
    """Raised when a message is not found."""

    def __init__(self, params: tuple = (), message: str = "Message not found.", code: int = 152):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidInvitationTicket(FruitCraftException):
    """Raised when an invitation ticket is invalid."""

    def __init__(self, params: tuple = (), message: str = "Invalid invitation ticket.", code: int = 153):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidRestoreKey(FruitCraftException):
    """Raised when a restore key is invalid."""

    def __init__(self, params: tuple = (),
                 message: str = "You have entered an invalid restore key. Please contact customer care for more information.",
                 code: int = 154):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AccountChangeLimit(FruitCraftException):
    """Raised when the account cannot be changed more than once a day."""

    def __init__(self, params: tuple = (), message: str = "You cannot change your account more than once a day.",
                 code: int = 155):
        super().__init__(message % params)
        self.code = code
        self.params = params


class IdentificationRequired(FruitCraftException):
    """Raised when identification information is required."""

    def __init__(self, params: tuple = (), message: str = "We need your name or restore key for your identification.",
                 code: int = 156):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerNotFound(FruitCraftException):
    """Raised when a player is not found."""

    def __init__(self, params: tuple = (),
                 message: str = "Player not found, please try again. If problem persists, be sure to let us know.",
                 code: int = 157):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvitationCodeAlreadyRedeemed(FruitCraftException):
    """Raised when an invitation code has already been redeemed."""

    def __init__(self, params: tuple = (), message: str = "You have already redeemed an invitation code.",
                 code: int = 158):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidEmail(FruitCraftException):
    """Raised when an invalid email address is provided."""

    def __init__(self, params: tuple = (), message: str = "Please enter a valid e-Mail address.", code: int = 159):
        super().__init__(message % params)
        self.code = code
        self.params = params


class EmailAlreadyExists(FruitCraftException):
    """Raised when the email address already exists."""

    def __init__(self, params: tuple = (), message: str = "e-Mail already exists.", code: int = 160):
        super().__init__(message % params)
        self.code = code
        self.params = params


class EmailAlreadyRegistered(FruitCraftException):
    """Raised when the email address has already been registered."""

    def __init__(self, params: tuple = (), message: str = "You have already registered a valid e-Mail address.",
                 code: int = 161):
        super().__init__(message % params)
        self.code = code
        self.params = params


class RewardAlreadyReceived(FruitCraftException):
    """Raised when a reward has already been received."""

    def __init__(self, params: tuple = (), message: str = "You have already received this reward.", code: int = 162):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AccountDeactivated(FruitCraftException):
    """Raised when the account is deactivated."""

    def __init__(self, params: tuple = (), message: str = "Your account is deactivated.", code: int = 163):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AccountResetFailed(FruitCraftException):
    """Raised when account reset fails."""

    def __init__(self, params: tuple = (), message: str = "Account reset failed.", code: int = 164):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidActivationCode(FruitCraftException):
    """Raised when the activation code is invalid."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "Your activation code is invalid. Please contact customer care for more information.",
                 code: int = 165):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CashRewardNotFound(FruitCraftException):
    """Raised when the specified cash reward cannot be found."""

    def __init__(self, params: tuple = (), message: str = "The specified cash reward cannot be found.",
                 code: int = 166):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AvatarNotAvailable(FruitCraftException):
    """Raised when the selected avatar is not available."""

    def __init__(self, params: tuple = (), message: str = "Avatar is not available. Please try another one.",
                 code: int = 167):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotInLeague(FruitCraftException):
    """Raised when the user is not in a league."""

    def __init__(self, params: tuple = (), message: str = "You are not in a league.", code: int = 168):
        super().__init__(message % params)
        self.code = code
        self.params = params


class DeviceVerificationError(FruitCraftException):
    """Raised when there is trouble verifying the device."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "We are having trouble to verify your device. Please contact customer care for more information.",
                 code: int = 169):
        super().__init__(message % params)
        self.code = code
        self.params = params


class StoreNotSupported(FruitCraftException):
    """Raised when the selected store is not supported."""

    def __init__(self, params: tuple = (), message: str = "The selected store is not supported yet.", code: int = 170):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PackInconsistencyError(FruitCraftException):
    """Raised for inconsistencies related to the selected pack."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "An inconsistency occurred about your selected pack. Please restart your game and try again.",
                 code: int = 171):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PaymentReceiptNotProvided(FruitCraftException):
    """Raised when the receipt for a payment is not provided."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "The receipt for your payment is not provided. If the problem persists, please contact customer care.",
                 code: int = 172):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardCannotEvolve(FruitCraftException):
    """Raised when a card cannot be evolved further."""

    def __init__(self, params: tuple = (), message: str = "You cannot evolve this card further.", code: int = 173):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InternalError(FruitCraftException):
    """Raised for internal errors."""

    def __init__(self, params: tuple = (), message: str = "An internal error occurred. We are trying to fix it ASAP.",
                 code: int = 174):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InternalErrorWait(FruitCraftException):
    """Raised for internal errors requiring a wait."""

    def __init__(self, params: tuple = (),
                 message: str = "An internal error occurred. Please wait for a moment and try again.",
                 code: int = 175):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerNotParticipated(FruitCraftException):
    """Raised when a player did not participate in a battle."""

    def __init__(self, params: tuple = (), message: str = "Player did not participate in this battle.",
                 code: int = 176):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardCoolingDown(FruitCraftException):
    """Raised when a card is cooling down and unavailable."""

    def __init__(self, params: tuple = (), message: str = "Card is cooling down and unavailable right now.",
                 code: int = 177):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CardCoolEnough(FruitCraftException):
    """Raised when a card is cool enough to be used."""

    def __init__(self, params: tuple = (), message: str = "Card is cool enough.", code: int = 178):
        super().__init__(message % params)
        self.code = code
        self.params = params


class DeviceInconsistency(FruitCraftException):
    """Raised when there is an inconsistency about the device."""

    def __init__(self, params: tuple = (),
                 message: str = "There is an inconsistency about your device. Please contact customer care for more information.",
                 code: int = 179):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NameTooLong(FruitCraftException):
    """Raised when the provided name is too long."""

    def __init__(self, params: tuple = (),
                 message: str = "Name is too long. You can specify %s characters for your name.",
                 code: int = 180):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NameAlreadyTaken(FruitCraftException):
    """Raised when the provided name is already taken."""

    def __init__(self, params: tuple = (), message: str = "Name is already taken.", code: int = 181):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CountryCodeNotSupported(FruitCraftException):
    """Raised when the provided country code is not supported."""

    def __init__(self, params: tuple = (), message: str = "Your selected country code is not supported yet.",
                 code: int = 182):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotEnoughGold(FruitCraftException):
    """Raised when there is not enough gold."""

    def __init__(self, params: tuple = (), message: str = "You need %s more gold.", code: int = 183):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OnlineOnAnotherDevice(FruitCraftException):
    """Raised when the player is online on another device."""

    def __init__(self, params: tuple = (), message: str = "You are currently online on another device.",
                 code: int = 184):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeNameTooLong(FruitCraftException):
    """Raised when the tribe name is too long."""

    def __init__(self, params: tuple = (), message: str = "Tribe name cannot exceed 30 characters.", code: int = 185):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TopTribeEditLimit(FruitCraftException):
    """Raised when the top tribe edit limit is reached."""

    def __init__(self, params: tuple = (), message: str = "Top 25 tribes cannot edit their information.",
                 code: int = 186):
        super().__init__(message % params)
        self.code = code
        self.params = params


class CannotSellCard(FruitCraftException):
    """Raised when a card cannot be sold."""

    def __init__(self, params: tuple = (), message: str = "You cannot sell crystal or hero cards.", code: int = 187):
        super().__init__(message % params)
        self.code = code
        self.params = params


class StatusRequired(FruitCraftException):
    """Raised when a status is required."""

    def __init__(self, params: tuple = (), message: str = "You need to specify a status!", code: int = 188):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeNotFound(FruitCraftException):
    """Raised when a tribe is not found."""

    def __init__(self, params: tuple = (), message: str = "Tribe not found.", code: int = 189):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NameRequired(FruitCraftException):
    """Raised when a name is required."""

    def __init__(self, params: tuple = (), message: str = "Name Required.", code: int = 190):
        super().__init__(message % params)
        self.code = code
        self.params = params


class DescriptionRequired(FruitCraftException):
    """Raised when a description is required."""

    def __init__(self, params: tuple = (), message: str = "Description Required.", code: int = 191):
        super().__init__(message % params)
        self.code = code
        self.params = params


class DescriptionTooLong(FruitCraftException):
    """Raised when the description is too long."""

    def __init__(self, params: tuple = (), message: str = "Description is Too Long.", code: int = 192):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeNameExists(FruitCraftException):
    """Raised when a tribe name already exists."""

    def __init__(self, params: tuple = (), message: str = "A tribe with the same name already exists.",
                 code: int = 193):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidChiefPlayers(FruitCraftException):
    """Raised when invalid chief players are provided."""

    def __init__(self, params: tuple = (), message: str = "Invalid chief players for tribe id %d", code: int = 194):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AlreadyInTribe(FruitCraftException):
    """Raised when a player is already a member of a tribe."""

    def __init__(self, params: tuple = (), message: str = "You are already a member of this tribe.", code: int = 195):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NoTribeBuilding(FruitCraftException):
    """Raised when the player does not have the tribe building yet."""

    def __init__(self, params: tuple = (), message: str = "You do not have the tribe building yet.", code: int = 196):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxTribeChangeLimit(FruitCraftException):
    """Raised when the maximum tribe change limitation is reached."""

    def __init__(self, params: tuple = (), message: str = "You have reached the maximum tribe change limitation.",
                 code: int = 197):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeNoMembers(FruitCraftException):
    """Raised when a tribe has no members."""

    def __init__(self, params: tuple = (), message: str = "Inconsistency exception! Tribe has no members.",
                 code: int = 198):
        super().__init__(message % params)
        self.code = code
        self.params = params


class UndecidedRequest(FruitCraftException):
    """Raised when there is an undecided request."""

    def __init__(self, params: tuple = (), message: str = "You already have an undecided request.", code: int = 199):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidDecisionParameter(FruitCraftException):
    """Raised when an invalid decision parameter is provided."""

    def __init__(self, params: tuple = (), message: str = "Invalid decision parameter. Please try again.",
                 code: int = 200):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotJoinRequest(FruitCraftException):
    """Raised when the request is not a join request."""

    def __init__(self, params: tuple = (),
                 message: str = "This is not a join request. Please contact customer care for more information.",
                 code: int = 201):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InconsistentDataProvided(FruitCraftException):
    """Raised when inconsistent data is provided."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "Inconsistent Data Provided, please try again. If problem persists, be sure to let us know.",
                 code: int = 202):
        super().__init__(message % params)
        self.code = code
        self.params = params


class JoinRequestProcessed(FruitCraftException):
    """Raised when a join request has already been processed."""

    def __init__(self, params: tuple = (), message: str = "Join request has already been processed.", code: int = 203):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeFull(FruitCraftException):
    """Raised when a tribe is full."""

    def __init__(self, params: tuple = (), message: str = "Tribe is full.", code: int = 204):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeAccessDenied(FruitCraftException):
    """Raised when tribe access permission is denied."""

    def __init__(self, params: tuple = (), message: str = "Tribe access permission denied.", code: int = 205):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerAlreadyInTribe(FruitCraftException):
    """Raised when a player is already a member of the tribe."""

    def __init__(self, params: tuple = (), message: str = "Player is already a member of this tribe.", code: int = 206):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotAnInvitation(FruitCraftException):
    """Raised when the request is not an invitation."""

    def __init__(self, params: tuple = (), message: str = "This is not an invitation.", code: int = 207):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvitationProcessed(FruitCraftException):
    """Raised when an invitation has already been processed."""

    def __init__(self, params: tuple = (), message: str = "Invitation has already been processed.", code: int = 208):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NoTribesAvailable(FruitCraftException):
    """Raised when there are no tribes available."""

    def __init__(self, params: tuple = (), message: str = "You do not have any tribes.", code: int = 209):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerNotInYourTribe(FruitCraftException):
    """Raised when a player is not in your tribe."""

    def __init__(self, params: tuple = (), message: str = "Player is not in your tribe.", code: int = 210):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerAlreadyPromoted(FruitCraftException):
    """Raised when a player is already promoted."""

    def __init__(self, params: tuple = (), message: str = "Player is already promoted.", code: int = 211):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerNotElder(FruitCraftException):
    """Raised when a player is not an elder."""

    def __init__(self, params: tuple = (), message: str = "Player is not an elder.", code: int = 212):
        super().__init__(message % params)
        self.code = code
        self.params = params


class SelfPoke(FruitCraftException):
    """Raised when a player tries to poke themselves."""

    def __init__(self, params: tuple = (), message: str = "You can not poke yourself.", code: int = 213):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotAMemberOfTribe(FruitCraftException):
    """Raised when a player is not a member of the tribe."""

    def __init__(self, params: tuple = (), message: str = "You are not a member of this tribe.", code: int = 214):
        super().__init__(message % params)
        self.code = code
        self.params = params


class SelfKick(FruitCraftException):
    """Raised when a player tries to kick themselves."""

    def __init__(self, params: tuple = (), message: str = "You can not kick yourself.", code: int = 215):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InsufficientPermission(FruitCraftException):
    """Raised when a player does not have enough permission to perform an operation."""

    def __init__(self, params: tuple = (), message: str = "You don't have enough permission to do this operation.",
                 code: int = 216):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxLevelCooldownBuilding(FruitCraftException):
    """Raised when the maximum level for the cooldown building is reached."""

    def __init__(self, params: tuple = (), message: str = "Maximum level reached for Cooldown building.",
                 code: int = 217):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxLevelMainHallBuilding(FruitCraftException):
    """Raised when the maximum level for the main hall building is reached."""

    def __init__(self, params: tuple = (), message: str = "Maximum level reached for Main Hall building.",
                 code: int = 218):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxLevelDefenseBuilding(FruitCraftException):
    """Raised when the maximum level for the defense building is reached."""

    def __init__(self, params: tuple = (), message: str = "Maximum level reached for Defense building.",
                 code: int = 219):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxLevelOffenseBuilding(FruitCraftException):
    """Raised when the maximum level for the offense building is reached."""

    def __init__(self, params: tuple = (), message: str = "Maximum level reached for Offense building.",
                 code: int = 220):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxLevelGoldBuilding(FruitCraftException):
    """Raised when the maximum level for the gold building is reached."""

    def __init__(self, params: tuple = (), message: str = "Maximum level reached for Gold building.", code: int = 221):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxLevelBankBuilding(FruitCraftException):
    """Raised when the maximum level for the bank building is reached."""

    def __init__(self, params: tuple = (), message: str = "Maximum level reached for Bank building.", code: int = 222):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidBuildingTypeForUpgrade(FruitCraftException):
    """Raised when an invalid building type is provided for upgrade."""

    def __init__(self, params: tuple = (), message: str = "Invalid building type for upgrade.", code: int = 223):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MinimumDonationAmount(FruitCraftException):
    """Raised when the minimum donation amount is not met."""

    def __init__(self, params: tuple = (), message: str = "Minimum donation amount is %s golds.", code: int = 224):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PlayerNoTribeBuilding(FruitCraftException):
    """Raised when a player does not have the tribe building."""

    def __init__(self, params: tuple = (), message: str = "Player does not have the tribe building yet.",
                 code: int = 225):
        super().__init__(message % params)
        self.code = code
        self.params = params


class UndecidedInvitation(FruitCraftException):
    """Raised when a user has an undecided invitation."""

    def __init__(self, params: tuple = (), message: str = "User has an undecided invitation already.", code: int = 226):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotEnoughTribeMoney(FruitCraftException):
    """Raised when there is not enough tribe money."""

    def __init__(self, params: tuple = (), message: str = "Not enough tribe money.", code: int = 227):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeScoreUpdateFailed(FruitCraftException):
    """Raised when updating the tribe score fails."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "Failed to update tribe score, please try again. If problem persists, be sure to let us know.",
                 code: int = 228):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeNoIdentifier(FruitCraftException):
    """Raised when a tribe has no identifier."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "Tribe has no identifier, please try again. If problem persists, be sure to let us know.",
                 code: int = 229):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidBuildingTypeForCardCapacity(FruitCraftException):
    """Raised when an invalid building type is provided for card capacity."""

    def __init__(self, params: tuple = (), message: str = "Invalid building type for card capacity.", code: int = 230):
        super().__init__(message % params)
        self.code = code
        self.params = params


class OperationFailed(FruitCraftException):
    """Raised when an operation fails."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "Operation failed, please try again. If problem persists, be sure to let us know.",
                 code: int = 231):
        super().__init__(message % params)
        self.code = code
        self.params = params


class GooglePlayVerificationFailed(FruitCraftException):
    """Raised when Google Play verification fails."""

    def __init__(self, params: tuple = (), message: str = "Google Play verification failed.", code: int = 232):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidPurchaseState(FruitCraftException):
    """Raised when the purchase state is invalid."""

    def __init__(self, params: tuple = (), message: str = "purchaseState is invalid. (%s)", code: int = 233):
        super().__init__(message % params)
        self.code = code
        self.params = params


class DataReadingError(FruitCraftException):
    """Raised when there is a problem reading data from the server."""

    def __init__(self, params: tuple = (), message: str = "Problem reading data from server", code: int = 234):
        super().__init__(message % params)
        self.code = code
        self.params = params


class SibcheVerificationFailed(FruitCraftException):
    """Raised when Sibche verification fails."""

    def __init__(self, params: tuple = (), message: str = "Sibche verification failed.", code: int = 235):
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxBoostsLimit(FruitCraftException):
    """Raised when the maximum boosts limit is reached."""

    def __init__(self, params: tuple = (), message: str = "You cannot buy more boosts.", code: int = 236):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidCountryCode(FruitCraftException):
    """Raised when an invalid country code is provided."""

    def __init__(self, params: tuple = (), message: str = "Invalid country code.", code: int = 237):
        super().__init__(message % params)
        self.code = code
        self.params = params


class UserNotFound(FruitCraftException):
    """Raised when a user is not found."""

    def __init__(self, params: tuple = (), message: str = "User Not Found", code: int = 238):
        super().__init__(message % params)
        self.code = code
        self.params = params


class UserRecentlyPoked(FruitCraftException):
    """Raised when a user has been recently poked."""

    def __init__(self, params: tuple = (), message: str = "User has been recently poked.", code: int = 239):
        super().__init__(message % params)
        self.code = code
        self.params = params


class LeagueUpdateInProgress(FruitCraftException):
    """Raised when league update is in progress."""

    def __init__(self, params: tuple = (), message: str = "Updating league is in progress. Please wait.",
                 code: int = 240):
        super().__init__(message % params)
        self.code = code
        self.params = params


class FeatureNotImplemented(FruitCraftException):
    """Raised when a feature is not implemented yet."""

    def __init__(self, params: tuple = (), message: str = "This feature is not implemented yet :)", code: int = 241):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidLeagueId(FruitCraftException):
    """Raised when an invalid league ID is provided."""

    def __init__(self,
                 params: tuple = (),
                 message: str = "Operation failed (invalid league ID). Please contact customer care for more information.",
                 code: int = 242):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeHelpAlreadyInProgress(FruitCraftException):
    """Raised when tribe help is already in progress."""

    def __init__(self, params: tuple = (), message: str = "You should be fast! Your tribe mates are already helping",
                 code: int = 243):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NameRequired(FruitCraftException):
    """Raised when a name is required."""

    def __init__(self, params: tuple = (), message: str = "You should write a name", code: int = 244):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TutorialUpdateParameters(FruitCraftException):
    """Raised when tutorial updating requires more parameters."""

    def __init__(self, params: tuple = (), message: str = "Tutorial updating requires more parameters",
                 code: int = 245):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NoTribeAvailableToCoach(FruitCraftException):
    """Raised when there is no tribe available to coach."""

    def __init__(self, params: tuple = (), message: str = "No tribe available to coach, Try later", code: int = 246):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotEnoughNectar(FruitCraftException):
    """Raised when there is not enough nectar."""

    def __init__(self, params: tuple = (), message: str = "You need %s more nectar.", code: int = 247):
        super().__init__(message % params)
        self.code = code
        self.params = params


class HeroItemNotPurchased(FruitCraftException):
    """Raised when a hero item has not been purchased."""

    def __init__(self, params: tuple = (), message: str = "You've not bought this hero item.", code: int = 248):
        super().__init__(message % params)
        self.code = code
        self.params = params


class HeroItemAlreadyPurchased(FruitCraftException):
    """Raised when a hero item has already been purchased."""

    def __init__(self, params: tuple = (), message: str = "You've already bought this hero item.", code: int = 249):
        super().__init__(message % params)
        self.code = code
        self.params = params


class AllHeroesPurchased(FruitCraftException):
    """Raised when all heroes have been purchased."""

    def __init__(self, params: tuple = (), message: str = "You've bought all of the heroes!", code: int = 250):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotEnoughPotion(FruitCraftException):
    """Raised when there is not enough potion."""

    def __init__(self, params: tuple = (), message: str = "You have not enough potion.", code: int = 251):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidGiftCode(FruitCraftException):
    """Raised when an invalid gift code is entered."""

    def __init__(self, params: tuple = (), message: str = "Wrong gift code entered.", code: int = 252):
        super().__init__(message % params)
        self.code = code
        self.params = params


class GiftCodeAlreadyRedeemed(FruitCraftException):
    """Raised when a gift code has already been redeemed."""

    def __init__(self, params: tuple = (), message: str = "You've already redeemed this gift code", code: int = 253):
        super().__init__(message % params)
        self.code = code
        self.params = params


class GiftCodeExpired(FruitCraftException):
    """Raised when a gift code is expired."""

    def __init__(self, params: tuple = (), message: str = "Gift code is expired.", code: int = 254):
        super().__init__(message % params)
        self.code = code
        self.params = params


class HeroLevelRequirement(FruitCraftException):
    """Raised when a hero level requirement is not met."""

    def __init__(self, params: tuple = (), message: str = "You need at least one hero with level %s or above.",
                 code: int = 255):
        super().__init__(message % params)
        self.code = code
        self.params = params


class TribeEntryNotAllowed(FruitCraftException):
    """Raised when a player cannot enter a tribe."""

    def __init__(self, params: tuple = (), message: str = "You can't enter this tribe. Try another one.",
                 code: int = 256):
        super().__init__(message % params)
        self.code = code
        self.params = params


class LevelRequirement(FruitCraftException):
    """Raised when a level requirement is not met."""

    def __init__(self, params: tuple = (), message: str = "You should reach at least level %s !", code: int = 257):
        super().__init__(message % params)
        self.code = code
        self.params = params


class PrizeAlreadyReceived(FruitCraftException):
    """Raised when a prize has already been received."""

    def __init__(self, params: tuple = (), message: str = "You have got enough prize! Wait a little", code: int = 258):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidMobileNumber(FruitCraftException):
    """Raised when an invalid mobile number is entered."""

    def __init__(self, params: tuple = (), message: str = "You should enter a valid mobile number!", code: int = 259):
        super().__init__(message % params)
        self.code = code
        self.params = params


class InvalidVerificationCode(FruitCraftException):
    """Raised when an invalid verification code is entered."""

    def __init__(self, params: tuple = (), message: str = "Not a valid verification code! Try again", code: int = 260):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotSubscribed(FruitCraftException):
    """Raised when a user is not subscribed."""

    def __init__(self, params: tuple = (), message: str = "You are not subscribed! First subscribe then try again",
                 code: int = 261):
        super().__init__(message % params)
        self.code = code
        self.params = params


class NotCharged(FruitCraftException):
    """Raised when a user is not charged."""

    def __init__(self, params: tuple = (), message: str = "You are not charged! First get charged then try again",
                 code: int = 262):
        super().__init__(message % params)
        self.code = code
        self.params = params


class LevelRequirementAgain(FruitCraftException):
    """Raised when a level requirement is not met again."""

    def __init__(self, params: tuple = (), message: str = "You should reach at least level %s !", code: int = 263):
        print(params)
        super().__init__(message % params)
        self.code = code
        self.params = params


class MaxTribeBroadcastReached(FruitCraftException):
    """Raised when the maximum tribe broadcast limit is reached."""

    def __init__(self, params: tuple = (), message: str = "You reached max tribe broadcast! Wait a while and try again",
                 code: int = 264):
        super().__init__(message % params)
        self.code = code
        self.params = params


class BundlePurchaseError(FruitCraftException):
    """Raised when there is an error in purchasing a bundle."""

    def __init__(self, params: tuple = (),
                 message: str = "Error occurred in purchasing bundle. Please contact customer care.",
                 code: int = 265):
        super().__init__(message % params)
        self.code = code
        self.params = params



from os.path import exists
from json import load, dump


class Session:
    def __init__(self, client):
        self.client = client

    def doesSessionExist(self):
        return self.client.session_name and exists(f"{self.client.session_name}.fb")

    def loadSessionData(self):
        return load(open(f"{self.client.session_name}.fb", encoding="UTF-8"))

    def saveSession(self, playerInfo):
        data = {"restore_key": self.client.restore_key,
                "passport": self.client.passport,
                "udid": self.client.udid,
                "mobile_model": self.client.mobile_model,
                "player": {"id": playerInfo["id"], "name": playerInfo["name"], "invite_key": playerInfo["invite_key"]}}
        dump(data, open(f"{self.client.session_name}.fb", "w"), indent=4)



class HeroWithItems:
    def __init__(self, base_hero_id: int, left_base_item_ids: list = [], right_base_item_ids: list = []):
        self.base_hero_id = base_hero_id
        self.left_item_ids = left_base_item_ids
        self.right_base_item_ids = right_base_item_ids


from ..exceptions import *

exceptions = {
    None: UnknownError,
    100: GeneralError,
    101: AccountBlocked,
    102: CardNotFound,
    103: NoCardSelected,
    104: CardAlreadyInAuction,
    105: NotEnoughCards,
    106: InconsistencyError,
    107: CardNotInAuction,
    108: MaxBidReached,
    109: CannotBidOwnCard,
    110: AlreadyHighestBidder,
    111: AuctionClosed,
    112: CardTypeQuery,
    113: CardTypeNotFound,
    114: CardTypeQueryDuplicate,
    115: NotOwnerError,
    116: AccessDenied,
    117: ServerMaintenance,
    118: NoCardsSpecified,
    119: NoOpponentSpecified,
    120: BattleNotAvailable,
    121: CannotAttackSelf,
    122: PlayerProtected,
    123: CaptchaRequired,
    124: PlayingOnAnotherDevice,
    125: OpponentNotFound,
    126: OpponentOutOfRange,
    127: OpponentDefenseEmpty,
    128: InvalidCaptcha,
    129: CardInUse,
    130: CardMarkedForSacrifice,
    131: MaxPowerReached,
    132: MaxEvolutionsExceeded,
    133: EvolveLimitExceeded,
    134: BuildingNotFound,
    135: MaxCardsAssigned,
    136: LiveBattleUnavailable,
    137: OpponentOffline,
    138: AlreadyInLiveBattle,
    139: OpponentBusy,
    140: CannotAttackTribeMate,
    141: BattleIdRequired,
    142: InconsistencyErrorUpdate,
    143: OperationTimeout,
    144: PlayerUnavailable,
    145: BattleNotFound,
    146: AttackerNotFound,
    147: DefenderNotFound,
    148: TribeNotFound,
    149: NotInTribe,
    150: TribeHelpLimit,
    151: MessageSizeExceeded,
    152: MessageNotFound,
    153: InvalidInvitationTicket,
    154: InvalidRestoreKey,
    155: AccountChangeLimit,
    156: IdentificationRequired,
    157: PlayerNotFound,
    158: InvitationCodeAlreadyRedeemed,
    159: InvalidEmail,
    160: EmailAlreadyExists,
    161: EmailAlreadyRegistered,
    162: RewardAlreadyReceived,
    163: AccountDeactivated,
    164: AccountResetFailed,
    165: InvalidActivationCode,
    166: CashRewardNotFound,
    167: AvatarNotAvailable,
    168: NotInLeague,
    169: DeviceVerificationError,
    170: StoreNotSupported,
    171: PackInconsistencyError,
    172: PaymentReceiptNotProvided,
    173: CardCannotEvolve,
    174: InternalError,
    175: InternalErrorWait,
    176: PlayerNotParticipated,
    177: CardCoolingDown,
    178: CardCoolEnough,
    179: DeviceInconsistency,
    180: NameTooLong,
    181: NameAlreadyTaken,
    182: CountryCodeNotSupported,
    183: NotEnoughGold,
    184: OnlineOnAnotherDevice,
    185: TribeNameTooLong,
    186: TopTribeEditLimit,
    187: CannotSellCard,
    188: StatusRequired,
    189: TribeNotFound,
    190: NameRequired,
    191: DescriptionRequired,
    192: DescriptionTooLong,
    193: TribeNameExists,
    194: InvalidChiefPlayers,
    195: AlreadyInTribe,
    196: NoTribeBuilding,
    197: MaxTribeChangeLimit,
    198: TribeNoMembers,
    199: UndecidedRequest,
    200: InvalidDecisionParameter,
    201: NotJoinRequest,
    202: InconsistentDataProvided,
    203: JoinRequestProcessed,
    204: TribeFull,
    205: TribeAccessDenied,
    206: PlayerAlreadyInTribe,
    207: NotAnInvitation,
    208: InvitationProcessed,
    209: NoTribesAvailable,
    210: PlayerNotInYourTribe,
    211: PlayerAlreadyPromoted,
    212: PlayerNotElder,
    213: SelfPoke,
    214: NotAMemberOfTribe,
    215: SelfKick,
    216: InsufficientPermission,
    217: MaxLevelCooldownBuilding,
    218: MaxLevelMainHallBuilding,
    219: MaxLevelDefenseBuilding,
    220: MaxLevelOffenseBuilding,
    221: MaxLevelGoldBuilding,
    222: MaxLevelBankBuilding,
    223: InvalidBuildingTypeForUpgrade,
    224: MinimumDonationAmount,
    225: PlayerNoTribeBuilding,
    226: UndecidedInvitation,
    227: NotEnoughTribeMoney,
    228: TribeScoreUpdateFailed,
    229: TribeNoIdentifier,
    230: InvalidBuildingTypeForCardCapacity,
    231: OperationFailed,
    232: GooglePlayVerificationFailed,
    233: InvalidPurchaseState,
    234: DataReadingError,
    235: SibcheVerificationFailed,
    236: MaxBoostsLimit,
    237: InvalidCountryCode,
    238: UserNotFound,
    239: UserRecentlyPoked,
    240: LeagueUpdateInProgress,
    241: FeatureNotImplemented,
    242: InvalidLeagueId,
    243: TribeHelpAlreadyInProgress,
    244: NameRequired,
    245: TutorialUpdateParameters,
    246: NoTribeAvailableToCoach,
    247: NotEnoughNectar,
    248: HeroItemNotPurchased,
    249: HeroItemAlreadyPurchased,
    250: AllHeroesPurchased,
    251: NotEnoughPotion,
    252: InvalidGiftCode,
    253: GiftCodeAlreadyRedeemed,
    254: GiftCodeExpired,
    255: HeroLevelRequirement,
    256: TribeEntryNotAllowed,
    257: LevelRequirement,
    258: PrizeAlreadyReceived,
    259: InvalidMobileNumber,
    260: InvalidVerificationCode,
    261: NotSubscribed,
    262: NotCharged,
    263: LevelRequirementAgain,
    264: MaxTribeBroadcastReached,
    265: BundlePurchaseError,
}

errors = {
    100: "An unexpected error occurred, please try again. If problem persists, be sure to let us know.",
    101: "Your account is blocked.",
    102: "Card not found.",
    103: "You should choose at least one card!",
    104: "Card is already in auction.",
    105: "Not enough cards.",
    106: "An inconsistency error is occured about your action, please try again. If problem persists, be sure to let us know.",
    107: "Card not found in auctions.",
    108: "You cannot bid on this item, it's already reached the maximum price.",
    109: "You cannot bid on your own cards.",
    110: "You are already the highest bidder.",
    111: "Too late! Auction is closed.",
    112: "What kind of card should I look for?",
    113: "We cannot find the card type selected.",
    114: "What kind of card should I look for?",
    115: "You are not the owner. Cannot proceed.",
    116: "Access denied.",
    117: "Server is in maintenance. We will be available soon.",
    118: "You need to specify some cards for this operation.",
    119: "You need to specify an opponent for this operation.",
    120: "You cant start the battle right now, please try again. If problem persists, be sure to let us know.",
    121: "You cannot attack yourself.",
    122: "Player is protected by shield, wait a few hours before attacking them again.",
    123: "You have to enter captcha code to proceed.",
    124: "You are currently playing on another device, please close the game on other devices and try again in a few seconds.",
    125: "Opponent not found.",
    126: "Opponent not in your range.",
    127: "Opponent's defence deck is empty.",
    128: "Captcha is invalid.",
    129: "Card is used in one of your buildings and cannot be used here.",
    130: "Card is already marked for sacrifice.",
    131: "Card has reached its maximum power.",
    132: "Cannot evolve more than two cards.",
    133: "Only two cards of the same type can be evolved.",
    134: "We cannot find the selected building. Please update to the latest version and try again.",
    135: "You cannot assigned any more cards to this building.",
    136: "Live-battle is not available due to some inconsistencies. We will be available soon.",
    137: "Opponent is no longer online.",
    138: "Already in live-battle.",
    139: "Opponent is busy.",
    140: "You cannot attack your tribemate.",
    141: "Battle ID required.",
    142: "An inconsistency error is occured. Please update your game to the latest version and try again.",
    143: "The operation is timed out. Please try again.",
    144: "Player is not available right now. Please try again in a moment.",
    145: "Battle not found. Please try again.",
    146: "Attacker not found. Please try again.",
    147: "Defender not found. Please try again.",
    148: "Tribe not found. Please try again.",
    149: "You are not in a tribe.",
    150: "Your tribe cannot use help for not conforming to the tribe member limitation rule.",
    151: "Message size exceeded.",
    152: "Message not found.",
    153: "Invalid invitation ticket.",
    154: "You have entered an invalid restore key. Please contact customer care for more information.",
    155: "You cannot change your account more than once a day.",
    156: "We need your name or restore key for your identification.",
    157: "Player not found, please try again. If problem persists, be sure to let us know.",
    158: "You have already redeemed an invitation code.",
    159: "Please enter a valid e-Mail address.",
    160: "e-Mail already exists.",
    161: "You have already registered a valid e-Mail address.",
    162: "You have already received this reward.",
    163: "Your account is deactivated.",
    164: "Account reset failed.",
    165: "Your activation code is invalid. Please contact customer care for more information.",
    166: "The specified cash reward cannot be found.",
    167: "Avatar is not available. Please try another one.",
    168: "You are not in a league.",
    169: "We are having trouble to verify your device. Please contact customer care for more information.",
    170: "The selected store is not supported yet.",
    171: "An inconsistency occured about your selected pack. Please restart your game and try again.",
    172: "The receipt for your payment is not provided. If the problem persists, please contact customer care.",
    173: "You cannot evolve this card further.",
    174: "An internal error is occured. We are trying to fix it ASAP.",
    175: "An internal error is occured. Please wait for a moment and try again.",
    176: "Player did not participate in this battle.",
    177: "Card is cooling down and unavailable right now.",
    178: "Card is cool enough.",
    179: "There is an inconsistency about your device. Please contact customer care for more information.",
    180: "Name is too long. You can specify %s characters for your name.",
    181: "Name is already taken.",
    182: "Your selected country code is not supported yet.",
    183: "You need %s more gold.",
    184: "You are currently online on another device.",
    185: "Tribe name cannot exceed 30 characters.",
    186: "Top 25 tribes cannot edit their information.",
    187: "You cannot sell crystal or hero cards.",
    188: "You need to specify a status!",
    189: "Tribe not found.",
    190: "Name Required.",
    191: "Description Required.",
    192: "Description is Too Lang.",
    193: "A tribe with the same name already exists.",
    194: "Invalid chief players for tribe id %1",
    195: "You are already a member of this tribe.",
    196: "You do not have the tribe building yet.",
    197: "You have reached the maximum tribe change limitation.",
    198: "Inconsistency exception! Tribe has no members.",
    199: "You already have an undecided request.",
    200: "Invalid decision parameter. Please try again.",
    201: "This is not a join request. Please contact customer care for more information.",
    202: "Inconsistent Data Provided, please try again. If problem persists, be sure to let us know.",
    203: "Join request has already been processed.",
    204: "Tribe is full.",
    205: "Tribe access permission denied.",
    206: "Player is already a member of this tribe.",
    207: "This is not an invitation.",
    208: "Invitation has already been processed.",
    209: "You do not have any tribes.",
    210: "Player is not in your tribe.",
    211: "Player is already promoted.",
    212: "Player is not an elder.",
    213: "You can not poke yourself.",
    214: "You are not a member of this tribe.",
    215: "You can not kick yourself.",
    216: "You dont have enough permission to do this operation.",
    217: "Maximum level reached for Cooldown building.",
    218: "Maximum level reached for Main Hall building.",
    219: "Maximum level reached for Defense building.",
    220: "Maximum level reached for Offense building.",
    221: "Maximum level reached for Gold building.",
    222: "Maximum level reached for Bank building.",
    223: "Invalid building type for upgrade.",
    224: "Minimum donation amount is %s golds.",
    225: "Player does not have the tribe building yet.",
    226: "User have an undecided invitation already.",
    227: "Not enough tribe money.",
    228: "Failed to update tribe score, please try again. If problem persists, be sure to let us know.",
    229: "Tribe has no identifier, please try again. If problem persists, be sure to let us know.",
    230: "Invalid building type for card capacity.",
    231: "Operation faild, please try again. If problem persists, be sure to let us know.",
    232: "Google Play verification failed.",
    233: "purchaseState is invalid. (%s)",
    234: "Problem reading data from server",
    235: "Sibche verification failed.",
    236: "You cannot buy more boosts.",
    237: "Invalid country code.",
    238: "User Not Found",
    239: "User has been recently poked.",
    240: "Updating league is in progress. Please wait.",
    241: "This feature is not implemented yet :)",
    242: "Operation failed (invalid league ID). Please contact customer care for more information.",
    243: "You should be fast! Your tribe mates are already helping",
    244: "You should write a name",
    245: "Tutorial updating requires more parameters",
    246: "No tribe available to coach, Try later",
    247: "You need %s more nectar.",
    248: "You've not bought this hero item.",
    249: "You've already bought this hero item.",
    250: "You've bought all of the heroes!",
    251: "You have not enough potion.",
    252: "Wrong gift code entered.",
    253: "You've already redeemed this gift code",
    254: "Gift code is expired.",
    255: "You need at least one hero with level %s or above.",
    256: "You can't enter this tribe. Try another one.",
    257: "You should reach at least level %s !",
    258: "You have got enough prize! Wait a little",
    259: "You should enter a valid mobile number!",
    260: "Not a valid verification code! Try again",
    261: "You are not subscribed! First subscribe then try again",
    262: "You are not charged! First get charged then try again",
    263: "You should reach at least level %s !",
    264: "You reached max tribe broadcast! Wait a while and try again",
    265: "Error occured in purchasing bundle. Please contact customer care."
}

mobile_models = (
    "Samsung Galaxy S22 Ultra",
    "Samsung Galaxy S21 FE",
    "Samsung Galaxy A52",
    "Samsung Galaxy A72",
    "Google Pixel 6 Pro",
    "Google Pixel 6",
    "Google Pixel 5a",
    "OnePlus 9 Pro",
    "OnePlus 9",
    "OnePlus Nord 2",
    "Xiaomi Mi 11 Ultra",
    "Xiaomi Mi 11",
    "Xiaomi Redmi Note 11 Pro",
    "Xiaomi Poco F3",
    "Huawei P50 Pro",
    "Huawei Mate 40 Pro",
    "Huawei Nova 9",
    "Oppo Find X5 Pro",
    "Oppo Reno 7 Pro",
    "Motorola Edge 20 Pro",
    "Motorola Moto G Stylus 2022",
    "Sony Xperia 1 III",
    "Sony Xperia 5 III",
    "Nokia X20",
    "Nokia G50",
    "Realme GT 2 Pro",
    "Realme 8 Pro",
    "Vivo X70 Pro",
    "Vivo V21 5G",
    "Asus ROG Phone 5",
    "Asus Zenfone 8",
    "TCL 20 Pro 5G",
    "ZTE Axon 20 5G",
    "Honor 50 Pro",
    "Honor 50",
    "Infinix Zero 5G",
    "Lava Agni 5G",
    "Micromax IN Note 2"
)



class Genders:
    """
    The constants of this class can be used as an argument for the
    `gender` parameter in the `Client.changePlayerGender` method.
    """
    UNDEFINED = 0
    MALE = 1
    FEMALE = 2

class Moods:
    """
    The constants of this class can be used as an argument for the
    `mood_id` parameter in the `Client.changePlayerMood` method.
    """
    HAPPY = 1
    ANGRY = 2
    ASHAMED = 3
    BORED = 4
    CALM = 5
    CHEERFUL = 6
    COLD = 7
    PROUD = 8
    CONFUSED = 9
    CRANKY = 10
    CRAZY = 11
    CURIOUS = 12
    DEPRESSED = 13
    DISAPPOINTED = 14
    GOOD = 15
    HOPEFUL = 16
    HUNGRY = 17
    INDIFFERENT = 18
    APATHETIC = 19
    ALONE = 20

class AuctionCategories:
    """
    The constants of this class can be used as an argument for the
    `category` parameter in the `Client.searchAuctionCardsByFilters` method.
    """
    COMMON_CARDS = 0
    CHRISTMAS_CARDS = 1
    MONSTER_CARDS = 2

class AuctionPriceFilter:
    """
    The constants of this class can be used as an argument for the
    `price_filter` parameter in the `Client.searchAuctionCardsByFilters` method.
    """
    LOWEST_PRICE = 0
    HIGHEST_PRICE = 1

class BuildingTypes:
    """
    The constants of this class can be used as an argument for the
    `building_type` parameter in the `Client.assignCardsToBuilding` method.
    """
    GOLD_MINE = 1
    OFFENSE_BUILDING = 2
    DEFENSE_BUILDING = 3

class CardPackTypes:
    """
    The constants of this class can be used as an argument for the
    `card_pack_type` parameter in the `Client.buyCardPack` method.

    The contents of each pack are as follows:
    - BROWN_PACK: 2 Level 1 cards.
    - GREEN_PACK: 2 Level 2 cards.
    - YELLOW_PACK: 2 Level 3 cards.
    - RED_PACK: 2 Level 3 or 4 cards.
    - SILVER_PACK: 20 Level 1 cards and 10 Level 2 cards.
    - GOLD_PACK: 2 Level 4 or 5 cards.
    - PLATINUM_PACK: 2 Level 5 or 6 cards.
    - BLACK_PACK: 2 Level 6 or 7 cards.
    - MONSTER_PACK: 1 Super Powerful Monster card.
    - CRYSTAL_PACK: 1 Crystal card.
    - HERO_PACK: 1 Hero card of your choice from `HeroCardPackTypes`.
    """
    BROWN_PACK = 1
    GREEN_PACK = 2
    YELLOW_PACK = 3
    RED_PACK = 4
    SILVER_PACK = 5
    GOLD_PACK = 6
    PLATINUM_PACK = 7
    BLACK_PACK = 8
    MONSTER_PACK = 16
    CRYSTAL_PACK = 25
    HERO_PACK = 32

class HeroCardPackTypes:
    """
    The constants of this class can be used as an argument for the
    `hero_card_pack_type` parameter in the `Client.buyHeroCardPack` method.
    """
    XAKHMI = 415
    XEBELUS = 515
    HUSHIDAR = 615
    SIBILU = 715

class TribeCapabilities:
    """
    The constants of this class can be used as an argument for the
    `hero_card_pack_type` parameter in the `Client.buyHeroCardPack` method.
    """
    ATTACK_BONUS = 1002
    DEFENSE_BONUS = 1003
    RECOVERY_TIME = 1004
    TRIBE_CAPACITY = 1005

class TribeStatuses:
    """
    The constants of this class can be used as an argument for the
    `status` parameter in the `Client.editTribeInfo` and `Client.createTribe` methods.
    """
    OPEN = 1
    INVITE_ONLY = 2
    CLOSED = 3




from ..crypto import Encryption
from ..configs import exceptions
from json import dumps, loads
from urllib3 import PoolManager


class Network:
    def __init__(self, passport, base_url=None, time_out=None):
        self.crypto = Encryption()
        self.http = PoolManager()
        self.time_out = time_out or 10
        self.base_url = base_url or "http://iran.fruitcraft.ir"
        self.headers = {"Accept-Encoding": "gzip",
                        "Connection": "Keep-Alive",
                        "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
                        "Cookie": f"FRUITPASSPORT={passport};",
                        "Host": "iran.fruitcraft.ir",
                        "User-Agent": "Dalvik/2.1.0"
                        }

    def sendRequest(self, path: str, input: dict = {"client": "iOS"}, method: str = "POST", max_attempts: int = 1):
        url = f"{self.base_url}/{path}"
        data = f"edata={self.crypto.encrypt(dumps(input))}" if method == "POST" else None
        for _ in range(max_attempts):
            response = self.http.request(method=method,
                                         url=url,
                                         headers=self.headers,
                                         body=data,
                                         timeout=self.time_out)

            if response.headers.get("content-type", "").startswith("image/"):
                return response.data
            # Decode response data to UTF-8 and remove any BOM.
            response = response.data.decode("UTF-8").lstrip('\ufeff')
            if response.startswith("<!DOCTYPE"):
                raise exceptions[None]()
            try:
                response = loads(self.crypto.decrypt(response))
                if response.get('status'):
                    return response["data"]
            except exceptions as e:
                pass
        else:
            try:
                error_code = response["data"]["code"]
                error_args = tuple(response["data"]["arguments"])
            except Exception as e:
                raise exceptions[None]()
            raise exceptions[error_code](error_args)




from ..network import Network
from ..utils import Utils
from ..session import Session
from ..models import HeroWithItems
from ..enums import TribeStatuses
from json import dumps
from typing import List, Union


class Client:
    def __init__(self, session_name: str = None, restore_key: str = None, passport: str = None,
                 udid: str = None, base_url: str = None, time_out: int = None) -> None:
        self.session_name = session_name
        self.session = Session(self)
        self.mobile_model = None
        self.queue_number = None
        self.constant_version = "142"

        if self.session.doesSessionExist():
            self.session_data = self.session.loadSessionData()
            self.restore_key, self.passport, self.udid, self.mobile_model = (self.session_data[key] for key in
                                                                             ("restore_key", "passport", "udid",
                                                                              "mobile_model"))
        else:
            self.restore_key = restore_key
            self.passport = passport or Utils.generateRandomPassport()
            self.udid = udid or Utils.generateRandomUdid()
            self.mobile_model = Utils.getRandomMobileModel()
        self.sendRequest = Network(passport=self.passport, base_url=base_url, time_out=time_out).sendRequest

    def loadPlayer(self, save_session: bool = False, kochava_uid: str = None, appsflyer_uid: str = None,
                   store_type: str = None, metrix_uid: str = None, mobile_model: str = None, device_name: str = None,
                   **params):
        self.mobile_model = mobile_model or self.mobile_model
        params.update({
            "game_version": "1.10.10700",
            "udid": self.udid,
            "os_type": 2,
            "os_version": "11",
            "model": self.mobile_model,
            "device_name": device_name or "unknown",
            "store_type": store_type or "bazar"
        })
        if self.restore_key:
            params["restore_key"] = self.restore_key
        if metrix_uid:
            params["metrix_uid"] = metrix_uid
        if kochava_uid:
            params["kochava_uid"] = kochava_uid
        if appsflyer_uid:
            params["appsflyer_uid"] = appsflyer_uid

        response = self.sendRequest("player/load", params)
        self.restore_key, self.queue_number, self.constant_version = \
            response["restore_key"], response["q"], response["latest_constants_version"]

        if save_session:
            self.session.saveSession(response)
        return response

    def getConstants(self, mobile_model: str = None, constant_version: Union[str, int] = None, store_type: str = "bazar"):
        response = self.sendRequest("device/constants", {
            "game_version": "1.10.10700",
            "os_version": "11",
            "model": mobile_model or self.mobile_model,
            "constant_version": str(constant_version) or self.constant_version,
            "store_type": store_type
        })
        self.constant_version = response["LATEST_CONSTANTS_VERSION"]
        return response

    def getOpponents(self):
        return self.sendRequest("battle/getopponents")

    def applyInviteCode(self, code: str):
        return self.loadPlayer(invitation_ticket=code)

    def getPlayerInfo(self, player_id: int):
        return self.sendRequest("player/getplayerinfo", {"player_id": player_id})

    def setRecoveryEmail(self, email_address: str):
        return self.sendRequest("player/setemail", {"new_email": email_address})

    def setMobileNumber(self, mobile_number: str):
        return self.sendRequest("player/setmobilenumber", {"mobile_number": mobile_number})

    def setPlayerInfo(self, **params):
        return self.sendRequest("player/setplayerinfo", params)

    def changePlayerName(self, name: str):
        return self.setPlayerInfo(name=name, lang='fa')

    def changePlayerAvatar(self, avatar_id: int):
        return self.setPlayerInfo(avatar_id=avatar_id)

    def changePlayerMood(self, mood_id: int):
        return self.setPlayerInfo(mood_id=mood_id)

    def changePlayerAddress(self, address: str):
        return self.setPlayerInfo(address=address)

    def changePlayerPhone(self, phone: str):
        return self.setPlayerInfo(phone=phone)

    def changePlayerGender(self, gender: int):
        return self.setPlayerInfo(gender=gender)

    def changePlayerRealName(self, realname: str):
        return self.setPlayerInfo(realname=realname)

    def changePlayerBirthYear(self, birth_year: int):
        return self.setPlayerInfo(birth_year=birth_year)

    def claimAdReward(self):
        return self.sendRequest("player/claimadvertismentreward", {"check": Utils.hashQueueNumber(self.queue_number)})

    def getRecommendedTribes(self):
        return self.sendRequest("tribe/recommended")

    def createTribe(self, name: str, description: str, status: int = TribeStatuses.CLOSED):
        return self.sendRequest("tribe/create", {
            "name": name,
            "description": description,
            "status": status
        })

    def donateTribe(self, tribe_id: int, gold_amount: int):
        return self.sendRequest("tribe/donate", {
            "tribe_id": tribe_id,
            "gold": gold_amount
        })

    def joinTribe(self, tribe_id: int):
        return self.sendRequest("tribe/joinrequest", {"tribe_id": tribe_id})

    def upgradeBank(self):
        return self.sendRequest("tribe/upgrade", {"type": 1007})

    def upgradeGoldMine(self):
        return self.sendRequest("tribe/upgrade", {"type": 1001})

    def upgradeTribeCapability(self, feature: int, tribe_id: int):
        return self.sendRequest("tribe/upgrade", {
            "type": feature,
            "tribe_id": tribe_id
        })

    def cooloffCard(self, card_id: int):
        return self.sendRequest("cards/cooloff", {"card_id": card_id})

    def buyHeroItem(self, item_id: int):
        return self.sendRequest("store/buyheroitem", {"id": item_id})

    def equipHeroesItems(self, heroes_data: List[HeroWithItems], default_base_hero_id: int = None):
        if isinstance(heroes_data, HeroWithItems):
            heroes_data = [heroes_data]
        heroes_info = {
            "hero_details": [
                {
                    "items": [
                                 {"position": 1, "base_heroitem_id": item_id} for item_id in hero.left_item_ids
                             ] + [
                                 {"position": -1, "base_heroitem_id": item_id} for item_id in hero.right_base_item_ids
                             ],
                    "hero_id": hero.base_hero_id
                } for hero in heroes_data
            ]
        }
        heroes_info["hero_details"] = dumps(heroes_info["hero_details"], separators=(",", ":"))
        if default_base_hero_id:
            heroes_info["default_hero_id"] = default_base_hero_id
        return self.sendRequest("cards/equipheroitems", heroes_info)

    def broadcastMessageToTribe(self, title: str, message: str):
        return self.sendRequest("tribe/broadcast", {
            "title": title,
            "message": message
        })

    def getTribeBroadcasts(self):
        return self.sendRequest("message/tribebroadcast")

    def editTribeInfo(self, tribe_id: int, status: int, name: str, description: str):
        return self.sendRequest("tribe/edit", {
            "tribe_id": tribe_id,
            "status": status,
            "name": name,
            "description": description
        })

    def invitePlayerToJoinTribe(self, tribe_id: int, player_name: str):
        return self.sendRequest("tribe/invite", {
            "tribe_id": tribe_id,
            "invitee_name": player_name
        })

    def respondToTribeInvite(self, tribe_id: int, request_id: int, accept: bool):
        return self.sendRequest("tribe/decideinvite", {
            "tribe_id": tribe_id,
            "decision": "approve" if accept else "reject",
            "req_id": request_id
        })

    def respondToMembershipRequest(self, tribe_id: int, new_member_id: int, request_id: int, accept: bool):
        return self.sendRequest("tribe/decidejoin", {
            "tribe_id": tribe_id,
            "new_member_id": new_member_id,
            "decision": "approve" if accept else "reject",
            "req_id": request_id
        })

    def promoteTribeMember(self, tribe_id: int, member_id: int):
        return self.sendRequest("tribe/promote", {
            "tribe_id": tribe_id,
            "member_id": member_id
        })

    def searchTribes(self, query: str):
        return self.sendRequest("tribe/find", {"query": query})

    def demoteTribeMember(self, tribe_id: int, member_id: int):
        return self.sendRequest("tribe/demote", {
            "tribe_id": tribe_id,
            "member_id": member_id
        })

    def kickTribeMember(self, tribe_id: int, member_id: int):
        return self.sendRequest("tribe/kick", {
            "tribe_id": tribe_id,
            "member_id": member_id
        })

    def leaveTribe(self):
        return self.sendRequest("tribe/leave")

    def getTribeMembers(self, coach_tribe: bool = False):
        return self.sendRequest("tribe/members", {"coach_tribe": coach_tribe})

    def collectMinedGold(self):
        return self.sendRequest("cards/collectgold")

    def getShopItems(self):
        return self.sendRequest("store/getshopitems")

    def searchAuctionCards(self, query_type: int, **params):
        params["query_type"] = query_type
        return self.sendRequest("auction/search", params)

    def searchAuctionCardsByPower(self):
        return self.searchAuctionCards(1)

    def searchAuctionCardsByTime(self):
        return self.searchAuctionCards(3)

    def searchAuctionCardsById(self, base_card_id: int):
        return self.searchAuctionCards(2, base_card_id=base_card_id)

    def searchAuctionCardsByFilters(self, category: int, price_filter: int, card_level: int):
        return self.searchAuctionCards(6, category=category, cheapest=price_filter, rarity=card_level)

    def buyCardPack(self, card_pack_type: int, **params):
        params["type"] = card_pack_type
        return self.sendRequest("store/buycardpack", params)

    def buyHeroCardPack(self, hero_card_pack_type: int):
        return self.buyCardPack(32, base_card_id=hero_card_pack_type)

    def buyBoostPack(self, type: int, with_nectar: bool = True):
        return self.sendRequest("store/buyboostpack", {"with_nectar": with_nectar, "type": type})

    def buyAvatarPack(self):
        return self.sendRequest("store/buyavatarpack")

    def buyPotion(self, amount: int = None):
        return self.sendRequest("player/fillpotion", {"amount": amount} if amount != None else [])

    def potionizeHero(self, amount: int, base_hero_id: int):
        return self.sendRequest("cards/potionize", {
            "potion": amount,
            "hero_id": base_hero_id
        })

    def fullyPotionizeHeroByNectar(self, base_hero_id: int):
        return self.sendRequest("cards/potionize", {"hero_id": base_hero_id})

    def reportMessage(self, message_text: str, player_name: str, message_sent_date: int):
        return self.sendRequest("tribe/reportchat", {
            "message": message_text,
            "player_name": player_name,
            "date": message_sent_date
        })

    def getEventMessages(self):
        return self.sendRequest("message/systemmessages")

    def attackOpponent(self, opponent_id: int, card_ids: list, hero_id: int = None, number_attacks_today: int = 0):
        input = {"opponent_id": opponent_id,
                 "check": Utils.hashQueueNumber(self.queue_number),
                 "attacks_in_today": number_attacks_today}

        if hero_id:
            input["hero_id"] = hero_id
            if hero_id not in card_ids:
                card_ids.append(hero_id)
        input["cards"] = dumps(card_ids, separators=(',', ':'))
        response = self.sendRequest("battle/battle", input)
        if response.get("q"):
            self.queue_number = response.get("q")
        return response

    def doQuest(self, card_ids: list, hero_id: int = None):
        input = {"check": Utils.hashQueueNumber(self.queue_number)}
        if hero_id:
            input["hero_id"] = hero_id
            if hero_id not in card_ids:
                card_ids.append(hero_id)
        input["cards"] = dumps(card_ids, separators=(',', ':'))
        response = self.sendRequest("battle/quest", input)
        return response

    def startLiveBattle(self, opponent_id: int):
        return self.sendRequest("live-battle/livebattle", {"opponent_id": opponent_id})

    def joinLiveBattle(self, main_enemy_id: int, battle_id: int):
        return self.sendRequest("live-battle/livebattlejoin", {
            "mainEnemy": main_enemy_id,
            "battle_id": battle_id
        })

    def ackLiveBattleRequest(self, battle_id: int, already_in_game: bool = False):
        return self.sendRequest("live-battle/livebattleack", {
            "battle_id": battle_id,
            "already_in_game": str(int(already_in_game))
        })

    def respondToAttackRequest(self, accept: bool, battle_id: int):
        return self.sendRequest("live-battle/livebattlechoos", {
            "choice": int(accept),
            "battle_id": battle_id
        })

    def setCardForLiveBattle(self, battle_round: int, card_id: int, battle_id: int):
        return self.sendRequest("live-battle/setcardforlivebattle", {
            "round": battle_round,
            "card": card_id,
            "battle_id": battle_id
        })

    def invokeHeroAbilityInLiveBattle(self, battle_id: int, hero_id: int, ability_type: int):
        return self.sendRequest("live-battle/triggerability", {
            "hero_id": hero_id,
            "ability_type": ability_type,
            "battle_id": battle_id
        })

    def alertTribeToHelpInBattle(self, battle_id: int):
        return self.sendRequest("live-battle/help", {"battle_id": battle_id})

    def withdrawFromBank(self, amount: int):
        return self.sendRequest("player/withdrawfrombank", {"amount": amount})

    def depositToBank(self, amount: int):
        return self.sendRequest("player/deposittobank", {"amount": amount})

    def submitCardForAuction(self, card_id: int):
        return self.sendRequest("auction/setcardforauction", {"card_id": card_id})

    def getMyAuctionBidsHistory(self):
        return self.sendRequest("auction/loadmyparticipatedauctions")

    def getMyAuctionSalesHistory(self):
        return self.sendRequest("auction/loadmyauctions")

    def bidUpCardInAuction(self, auction_id: int):
        return self.sendRequest("auction/bid", {"auction_id": auction_id})

    def sellCardNow(self, auction_id: int):
        return self.sendRequest("auction/sellnow", {"auction_id": auction_id})

    def saveCardInAuctionBidsHistory(self, auction_id: int):
        return self.sendRequest("auction/follow", {"auction_id": auction_id})

    def setOnlineStatus(self, visible: bool = True):
        return self.sendRequest("tribe/invisible", {"status": 1 if visible else 3})

    def pokeTribeMember(self, tribe_id: int, member_id: int):
        return self.sendRequest("tribe/poke", {
            "tribe_id": tribe_id,
            "member_id": member_id
        })

    def openChanceBox(self):
        return self.sendRequest("player/turnthewheel",{
            "store": "4",
            "check": Utils.hashQueueNumber(self.queue_number)
        })

    def assignCardsToBuilding(self, card_ids: list, building_type: int):
        return self.sendRequest("cards/assign", {
            "cards": dumps(card_ids, separators=(",", ":")),
            "type": building_type
        })

    def getMyLeagueRanking(self):
        return self.sendRequest("ranking/league")

    def getTribeRankingBasedOnXP(self):
        return self.sendRequest("ranking/tribe")

    def getTribeRankingBasedOnSeed(self):
        return self.sendRequest("ranking/tribebasedonseed")

    def getGlobalRanking(self):
        return self.sendRequest("ranking/global")

    def getLocalRanking(self):
        return self.sendRequest("ranking/local")

    def getRisingRanking(self):
        return self.sendRequest("ranking/rising")

    def getLeaguesHistory(self, target_weeks: list):
        return self.sendRequest("ranking/leaguehistory", {"rounds": str(target_weeks)})

    def evolveCard(self, sacrifice_card_ids: list):
        return self.sendRequest("cards/evolve", {"sacrifices": str(sacrifice_card_ids)})

    def evolveHeroCard(self, hero_card_id: int):
        return self.evolveCard([hero_card_id])

    def enhanceCard(self, card_id: int, sacrifice_card_ids: list):
        return self.sendRequest("cards/enhance", {
            "card_id": card_id,
            "sacrifices": str(sacrifice_card_ids)
        })

    def redeemGiftCode(self, gift_code: str):
        return self.sendRequest("player/redeemgift", {"code": gift_code})

    def comebackToGame(self):
        return self.sendRequest("player/comeback")

    def scoutBattle(self, do_you_know_this_method_is_useless: bool):
        return self.sendRequest("battle/scout")

    def getMyCollection(self):
        return self.sendRequest("cards/collection")

    def claimFacebookReward(self):
        return self.sendRequest("player/claimfbreward")

    def claimInstagramReward(self):
        return self.sendRequest("player/claiminstagramreward")

    def getAllCardsInfo(self):
        return self.sendRequest("cards/cardsjsonexport", method="GET")

    def getAllFruitsInfo(self):
        return self.sendRequest("cards/fruitsjsonexport")

    def getCaptcha(self) -> bytes:
        return self.sendRequest("bot/getcaptcha", method="GET")



from ..configs import mobile_models
from random import choice
from string import ascii_lowercase, digits
from hashlib import md5

class Utils:
    def generateRandomPassport():
        return "".join(choice(ascii_lowercase + digits) for _ in range(32))
    def generateRandomUdid():
        return "".join(choice(ascii_lowercase + digits) for _ in range(16))
    def getRandomMobileModel():
        return choice(mobile_models)
    def hashQueueNumber(q: int):
        return md5(str(q).encode()).hexdigest()


from setuptools import setup, find_packages

with open("README.md", encoding="UTF-8") as f:
    readme = f.read()

setup(
    name="fruitbot",
    version="1.0.1b2",
    packages=find_packages(),
    install_requires=["urllib3"],
    author="Amir S.Farahani",
    url="https://github.com/AmirSF01/fruitcraft",
    description="A simple, easy-to-use and powerful library for automating gameplay in the Fruit Craft game through APIs.",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Games/Entertainment",
        "Topic :: Games/Entertainment :: Real Time Strategy",
    ],
    keywords=["fruit", "craft", "fruitcraft", "fruitbot", "bot", "api", "client", "automating", "card", "game"],
    python_requires=">=3.7"
)